﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
// Jay (the Ham) Harland

namespace Mystical_Ninja_Compression
{
    class Program
    {
        static void Main(string[] args)
        {
            
            if (args.Length == 0)
            {
                Console.WriteLine("Mystical Ninja for N64 compression");
                Console.WriteLine("By JaytheHam");
                Console.WriteLine("Drag and drop a compressed file to decompress it");
                Console.Read();
                return;
            }
            

            //string file = Console.ReadLine();

            FileStream inputFile = new FileStream(args[0], FileMode.Open);
            byte[] compFile = new byte[inputFile.Length];
            inputFile.Read(compFile, 0, compFile.Length);
            inputFile.Close();

            byte[] decompFile = Decompress(compFile, 0);

            FileStream outputFile = new FileStream("output", FileMode.Create);
            outputFile.Write(decompFile, 0, (int)decompFile.Length);
            outputFile.Flush();
            outputFile.Close();


        }

        /* Given a byte array containing a compressed Mystical Ninja file and the offset
         * within that file where the compressed file begins this returns a byte
         * array containing the decompressed file.
         */
        public static byte[] Decompress(byte[] inputFile, int startOffset)
        {
            uint compFileSize = (uint)
                ((inputFile[startOffset] << 24) +
                (inputFile[startOffset + 1] << 16) +
                (inputFile[startOffset + 2] << 8) +
                inputFile[startOffset + 3]);

            List<byte> decompFile = new List<byte>();

            uint readPosition = (uint)startOffset + 4;

            byte command = inputFile[readPosition];
            readPosition++;

            while (readPosition < (compFileSize + startOffset))
            {
                if (command < 0x80)
                {
                    //Copy already decompressed bytes

                    byte command2 = inputFile[readPosition];
                    readPosition++;

                    //Sliding window of size 0x3FF
                    int lookBack = ((command << 8) + command2) & 0x3FF;
                    lookBack = decompFile.Count - lookBack;

                    int toCopy = (command >> 2) + 2;

                    while (toCopy > 0)
                    {
                        decompFile.Add(decompFile[lookBack]);
                        lookBack++;
                        toCopy--;
                    }

                }
                else if (command < 0xA0)
                {
                    //Raw data copy

                    int toCopy = command & 0x1F;

                    while (toCopy > 0)
                    {
                        decompFile.Add(inputFile[readPosition]);
                        readPosition++;
                        toCopy--;
                    }
                }
                else if (command < 0xE0)
                {
                    //Write a specific repeditive byte

                    int toCopy = (command & 0x1F) + 2;
                    byte value = inputFile[readPosition];
                    readPosition++;

                    while (toCopy > 0)
                    {
                        decompFile.Add(value);
                        toCopy--;
                    }
                }
                else if (command < 0xFF)
                {
                    //Write 00s

                    int toCopy = (command & 0x1F) + 2;

                    while (toCopy > 0)
                    {
                        decompFile.Add(0x00);
                        toCopy--;
                    }
                }
                else //Command is 0xFF
                {
                    //Write a whole load of zeros
                    byte command2 = inputFile[readPosition];
                    readPosition++;

                    int toCopy = command2 + 2;

                    while (toCopy > 0)
                    {
                        decompFile.Add(0x00);
                        toCopy--;
                    }

                }

                command = inputFile[readPosition];
                readPosition++;
                
            }


            return (decompFile.ToArray());
        }
    }
}
